package at.kugel.plugin.dependency;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.internal.ui.javaeditor.ICompilationUnitDocumentProvider;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IPartListener2;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchPartReference;
import org.eclipse.ui.part.ViewPart;
import org.eclipse.ui.texteditor.IDocumentProvider;
import org.eclipse.ui.texteditor.ITextEditor;

/**
 * Workbench view plugin to show all used types, i.e. all outgoing references (even from
 * same package) of the current Java source that has the focos.
 * @author Kugel, <i>Theossos Comp Group</i>
 */
public final class OutgoingReferencesView extends ViewPart implements ISelectionListener
{

   // --- being a workbench part ---

   /**
    * @see IWorkbenchPart#createPartControl
    */
   @Override
   public void createPartControl(Composite parent)
   {
      // create the viewer and initialize it
      usedTypesTable = new TableViewer(parent, /* SWT.MULTI | */SWT.H_SCROLL | SWT.V_SCROLL);
      usedTypesTable.setContentProvider(new UsedTypesContentProvider());
      usedTypesTable.setLabelProvider(typesLabels);
      usedTypesTable.setInput(getViewSite());

      getSite().getWorkbenchWindow().getPartService().addPartListener(viewVisibleOrHiddenListener);

      hookDoubleClickAction();
   }

   /**
    * Passing the focus request to the viewer's control.
    */
   @Override
   public void setFocus()
   {
      usedTypesTable.getControl().setFocus();
   }

   /**
    * @see IWorkbenchPart#dispose
    */
   @Override
   public void dispose()
   {
      if (usedTypesTable != null)
      {
         getSite().getWorkbenchWindow().getPartService().removePartListener(viewVisibleOrHiddenListener);
         usedTypesTable = null;
      }
      super.dispose();
   }

   // --- model for view ---

   /**
    * The view shows data obtained from the model.
    */
   private TableViewer usedTypesTable;
   private String[] usedTypeModel = new String[0];
   private final ViewLabelProvider typesLabels = new ViewLabelProvider();

   /**
    * Provide objects to the view. A simple wrapper around the model.
    */
   class UsedTypesContentProvider implements IStructuredContentProvider
   {
      public void inputChanged(Viewer v, Object oldInput, Object newInput)
      {
      }

      public void dispose()
      {
      }

      public Object[] getElements(Object parent)
      {
         return usedTypeModel;
      }
   }

   // --- selection listening ---

   /**
    * Listener to be shown or hidden to start or stop listening for selections. Copied
    * from AbstractInfoView.<br />
    * @see org.eclipse.jdt.internal.ui.infoviews.AbstractInfoView#fPartListener
    */
   private final IPartListener2 viewVisibleOrHiddenListener = new IPartListener2()
   {
      private boolean isMySite(IWorkbenchPartReference ref)
      {
         return ref.getId().equals(getSite().getId());
      }

      /**
       * If our view gets visible, get the selection of the current active part. Start to
       * listen for selection changes.
       */
      public void partVisible(IWorkbenchPartReference ref)
      {
         if (isMySite(ref))
         {
            final IWorkbenchPart activePart = ref.getPage().getActivePart();
            if (activePart != null)
            {
               selectionChanged(activePart, ref.getPage().getSelection());
            }
            startListeningForSelectionChanges();
         }
      }

      public void partHidden(IWorkbenchPartReference pRef)
      {
         if (isMySite(pRef))
         {
            stopListeningForSelectionChanges();
         }
      }

      public void partInputChanged(IWorkbenchPartReference pRef)
      {
      }

      public void partActivated(IWorkbenchPartReference pRef)
      {
      }

      public void partBroughtToTop(IWorkbenchPartReference pRef)
      {
      }

      public void partClosed(IWorkbenchPartReference pRef)
      {
      }

      public void partDeactivated(IWorkbenchPartReference pRef)
      {
      }

      public void partOpened(IWorkbenchPartReference pRef)
      {
      }
   };

   /**
    * Start to listen for selection changes.
    */
   protected void startListeningForSelectionChanges()
   {
      getSite().getPage().addPostSelectionListener(this);
   }

   /**
    * Stop to listen for selection changes.
    */
   protected void stopListeningForSelectionChanges()
   {
      getSite().getPage().removePostSelectionListener(this);
   }

   /**
    * @see ISelectionListener#selectionChanged
    */
   @SuppressWarnings("restriction")
   public void selectionChanged(IWorkbenchPart pPart, ISelection pSelection)
   {
      boolean resetNameOfSelection = true;
      try
      {
         if (pSelection == null || pSelection.isEmpty())
         {
            // if we have no selection return
            return;
         }

         if (pSelection instanceof IStructuredSelection)
         {
            // selection in the navigator or package explorer or ...
            final IStructuredSelection selection = (IStructuredSelection) pSelection;
            if (selection.size() != 1)
            {
               return;
            }
            final Object element = selection.getFirstElement();
            if (element instanceof ICompilationUnit)
            {
               // compilation units e.g. in navigator
               resetNameOfSelection = !workOnAST((ICompilationUnit) element);
            }
            else if (element instanceof IType)
            {
               // source units e.g. in hierarchy
               resetNameOfSelection = !workOnAST(((IType) element).getCompilationUnit());
            }
            return;
         }

         if (pPart instanceof ITextEditor)
         {
            // selection in the Java or text editor
            final ITextEditor editor = (ITextEditor) pPart;
            if (editor.isDirty())
            {
               // only after save we update the view
               return;
            }
            final IDocumentProvider provider = editor.getDocumentProvider();
            // IDocument document = provider.getDocument(editor.getEditorInput());
            if (provider instanceof ICompilationUnitDocumentProvider)
            {
               resetNameOfSelection = !workOnAST(((ICompilationUnitDocumentProvider) provider).getWorkingCopy(editor.getEditorInput()));
            }
            return;
         }

      }
      finally
      {
         if (resetNameOfSelection)
         {
            setLastFullName(null);
         }
      }
   }

   // --- extract the information from AST ---

   private String lastFullName;

   private void setLastFullName(String lastFullName)
   {
      this.lastFullName = lastFullName;
      if (lastFullName != null)
      {
         typesLabels.setCurrentName(lastFullName);
      }
   }

   private boolean workOnAST(ICompilationUnit pCompilationUnit)
   {
      if (pCompilationUnit == null)
      {
         return false;
      }
      final String fullName = determineTypeName(pCompilationUnit);
      if (fullName.equals(lastFullName))
      {
         // do not work again on same type
         return true;
      }
      setLastFullName(fullName);
      // showMessage(fullName);

      // TODO do a background search for all incoming references as well

      final CompilationUnit syntaxTree = computeSyntaxTree(pCompilationUnit);
      final TypeCollectorVisitor astVisitor = new TypeCollectorVisitor();
      syntaxTree.accept(astVisitor);

      usedTypeModel = astVisitor.getUsedTypes();
      usedTypesTable.refresh();
      return true;
   }

   private String determineTypeName(ICompilationUnit compilationUnit)
   {
      final IType type = compilationUnit.findPrimaryType();
      final String elementName = type.getElementName();
      final String packageName = type.getPackageFragment().getElementName();
      return packageName != null ? packageName + "." + elementName : elementName;
   }

   private CompilationUnit computeSyntaxTree(ICompilationUnit compilationUnit)
   {
      // Eclipse 2.0 API CompilationUnit syntaxTree = AST.parseCompilationUnit(, true);
      final ASTParser c = ASTParser.newParser(AST.JLS3);
      c.setSource(compilationUnit);
      c.setResolveBindings(true);
      final ASTNode result = c.createAST(null);
      // showMessage(result.toString());
      final CompilationUnit syntaxTree = (CompilationUnit) result;
      return syntaxTree;
   }

   // --- click the view ---

   // private Action doubleClickAction;
   private void hookDoubleClickAction()
   {
      usedTypesTable.addDoubleClickListener(new IDoubleClickListener()
      {
         public void doubleClick(DoubleClickEvent event)
         {
            // showMessage("click " + event.getSelection().toString());
            // TODO open the type in the JDT editor
         }
      });
   }

   //   private void showMessage(String message)
   //   {
   //      MessageDialog.openInformation(usedTypesTable.getControl().getShell(), "Dependency View", message);
   //   }
   //
   //   private void logException(Exception e)
   //   {
   //      ILog logger = Activator.getDefault().getLog();
   //      IStatus status = new Status(IStatus.WARNING, Activator.PLUGIN_ID, IStatus.OK, e.getMessage(), e);
   //      logger.log(status);
   //   }

}
